//== GenVISC.h - Header file for "LLVM IR to VISC IR Pass" =//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#include "llvm/IR/Module.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/IntrinsicInst.h"
#include "llvm/IR/Value.h"
#include "llvm/Pass.h"
#include "llvm/IR/DerivedTypes.h"
#include "SupportVISC/VISCTimer.h"

using namespace llvm;

namespace genvisc {
// GenVISC - The first implementation.
struct GenVISC : public ModulePass {
  static char ID; // Pass identification, replacement for typeid
  GenVISC() : ModulePass(ID) {}


private:
  // Member variables
  Module* M;
  FunctionCallee llvm_visc_initializeTimerSet;
  FunctionCallee llvm_visc_switchToTimer;
  FunctionCallee llvm_visc_printTimerSet;

  GlobalVariable* TimerSet;

  // Functions
  void initializeTimerSet(Instruction*);
  void switchToTimer(enum visc_TimerID, Instruction*);
  void printTimerSet(Instruction*);
  Value* getStringPointer(const Twine& S, Instruction* InsertBefore, const Twine& Name = "");

public:
  // Functions
  virtual bool runOnModule(Module &M);

  void generateTest(CallInst* CI);
  Function* genKernel(Function* KernelFunction, CallInst* CI, StructType* RetTy);
  void genHost(CallInst*, Function*, unsigned, unsigned, unsigned, unsigned, StructType*);
};

} // End of namespace

