#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <string>
#include "kmeans.h"

#include <parboil.h>
#include <visc.h>

#ifdef WIN
#include <windows.h>
#else
#include <pthread.h>
#include <sys/time.h>
double gettime() {
    struct timeval t;
    gettimeofday(&t,NULL);
    return t.tv_sec+t.tv_usec*1e-6;
}
#endif


#ifdef NV
#else
#endif

#ifndef FLT_MAX
#define FLT_MAX 3.40282347e+38
#endif

#ifdef RD_WG_SIZE_0_0
#define BLOCK_SIZE RD_WG_SIZE_0_0
#elif defined(RD_WG_SIZE_0)
#define BLOCK_SIZE RD_WG_SIZE_0
#elif defined(RD_WG_SIZE)
#define BLOCK_SIZE RD_WG_SIZE
#else
#define BLOCK_SIZE 256
#endif

#ifdef RD_WG_SIZE_1_0
#define BLOCK_SIZE2 RD_WG_SIZE_1_0
#elif defined(RD_WG_SIZE_1)
#define BLOCK_SIZE2 RD_WG_SIZE_1
#elif defined(RD_WG_SIZE)
#define BLOCK_SIZE2 RD_WG_SIZE
#else
#define BLOCK_SIZE2 256
#endif


#ifndef FLT_MAX
#define FLT_MAX 3.40282347e+38
#endif

void
kmeans_kernel_c(float  *feature,
                float  *clusters,
                int    *membership,
                int     npoints,
                int     nclusters,
                int     nfeatures,
                int		offset,
                int		size
               )
{
    __visc__attributes(3, feature, clusters, membership, 1, membership);
    unsigned int point_id = get_global_id(0);
    int index = 0;
    //const unsigned int point_id = get_global_id(0);
    if (point_id < npoints)
    {
        float min_dist=FLT_MAX;
        for (int i=0; i < nclusters; i++) {

            float dist = 0;
            float ans  = 0;
            for (int l=0; l<nfeatures; l++) {
                ans += (feature[l * npoints + point_id]-clusters[i*nfeatures+l])*
                       (feature[l * npoints + point_id]-clusters[i*nfeatures+l]);
            }

            dist = ans;
            if (dist < min_dist) {
                min_dist = dist;
                index    = i;

            }
        }
        //printf("%d\n", index);
        membership[point_id] = index;
    }

}

void
kmeans_swap(float  *feature,
            float  *feature_swap,
            int     npoints,
            int     nfeatures
           ) {
    __visc__attributes(2, feature, feature_swap, 1, feature_swap);

    unsigned int tid = get_global_id(0);
    for(int i = 0; i <  nfeatures; i++)
        feature_swap[i * npoints + tid] = feature[tid * nfeatures + i];

}

// local variables


int   *membership_OCL;
int   *membership_d;
float *feature_d;
float *clusters_d;
float *center_d;

float* feature_swap_d;

static struct pb_TimerSet timers;

__attribute__((noinline)) int allocate(int n_points, int n_features, int n_clusters, float **feature)
{

    pb_InitializeTimerSet(&timers);
    __visc__init();
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

    size_t bytes = n_points * n_features * sizeof(float);
    feature_swap_d = (float*)malloc (bytes);
    pb_SwitchToTimer(&timers, visc_TimerID_MEM_TRACK);
    llvm_visc_track_mem(feature_swap_d, bytes);
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    unsigned global_work[3] = { n_points, 1, 1 };
    /// Ke Wang adjustable local group size 2013/08/07 10:37:33
    unsigned local_work_size= BLOCK_SIZE; // work group size is defined by RD_WG_SIZE_0 or RD_WG_SIZE_0_0 2014/06/10 17:00:51
    if(global_work[0]%local_work_size !=0)
        global_work[0]=(global_work[0]/local_work_size+1)*local_work_size;

    pb_SwitchToTimer(&timers, pb_TimerID_NONE);
    unsigned swapDFG = __visc__node(kmeans_swap, 2, 1, local_work_size,
        global_work[0]/local_work_size, 6, feature[0], bytes, (float*)
        feature_swap_d, bytes, n_points, n_features, 0);
    __visc__wait(swapDFG);

    //llvm_visc_request_mem(feature_swap_d, n_points * n_features * sizeof(float));
    //for(int i = 0; i < n_points * n_features; i++) {
      //std::cout << "feature_swap[" <<i << "] = " << feature_swap_d[i] << "\n";
    //}
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    membership_OCL = (int*) malloc(n_points * sizeof(int));
    pb_SwitchToTimer(&timers, visc_TimerID_MEM_TRACK);
    llvm_visc_track_mem(membership_OCL, n_points * sizeof(int));
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    return 0;
}

void deallocateMemory()
{
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    free(feature_swap_d);
    free(membership_OCL);
    pb_SwitchToTimer(&timers, visc_TimerID_MEM_UNTRACK);
    llvm_visc_untrack_mem(feature_swap_d);
    llvm_visc_untrack_mem(membership_OCL);
    pb_SwitchToTimer(&timers, pb_TimerID_NONE);
    pb_PrintTimerSet(&timers);
    __visc__cleanup();
}


int main( int argc, char** argv)
{
    
    printf("WG size of kernel_swap = %d, WG size of kernel_kmeans = %d \n", BLOCK_SIZE, BLOCK_SIZE2);

    setup(argc, argv);
}

__attribute__((noinline)) int kmeansOCL(float **feature,    /* in: [npoints][nfeatures] */
              int     n_features,
              int     n_points,
              int     n_clusters,
              int    *membership,
              float **clusters,
              int     *new_centers_len,
              float  **new_centers)
{

    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    int delta = 0;
    int i, j, k;

    unsigned global_work[3] = { n_points, 1, 1 };

    /// Ke Wang adjustable local group size 2013/08/07 10:37:33
    unsigned local_work_size=BLOCK_SIZE2; // work group size is defined by RD_WG_SIZE_1 or RD_WG_SIZE_1_0 2014/06/10 17:00:41
    if(global_work[0]%local_work_size !=0)
        global_work[0]=(global_work[0]/local_work_size+1)*local_work_size;

    int size = 0;
    int offset = 0;
    size_t bytes = n_clusters * n_features * sizeof(float);
    size_t bytes_feature = n_points * n_features * sizeof(float);
    size_t bytes_membership = n_points * sizeof(int);

    pb_SwitchToTimer(&timers, pb_TimerID_NONE);

    unsigned kmeansDFG = __visc__node(kmeans_kernel_c, 2, 1, local_work_size,
        global_work[0]/local_work_size, 11, feature_swap_d, bytes_feature, clusters[0],
        bytes, membership_OCL, bytes_membership, n_points, n_clusters, n_features,
        offset, size, 0);
    __visc__wait(kmeansDFG);
    pb_SwitchToTimer(&timers, pb_TimerID_COPY);

    llvm_visc_request_mem(membership_OCL, bytes_membership); // read membership into membership_OCL
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

    delta = 0;
    for (i = 0; i < n_points; i++)
    {
        int cluster_id = membership_OCL[i];
        new_centers_len[cluster_id]++;
        if (membership_OCL[i] != membership[i])
        {
            delta++;
            membership[i] = membership_OCL[i];
        }
        for (j = 0; j < n_features; j++)
        {
            new_centers[cluster_id][j] += feature[i][j];
        }
    }

    return delta;
}
