/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include "common.h"

__kernel void block2D_hybrid_coarsen_x(float c0,float c1,__global float* A0, __global float* Anext, int nx, int ny, int nz)
{
	//thread coarsening along x direction
	const int i = get_group_id(0)*get_local_size(0)*2+get_local_id(0);
	const int i2 = get_group_id(0)*get_local_size(0)*2+get_local_id(0)+get_local_size(0);
	const int j = get_group_id(1)*get_local_size(1)+get_local_id(1);
	
	//get available region for load and store
	const bool w_region =  i>0 && j>0 &&(i<nx-1) &&(j<ny-1) ;
	const bool w_region2 =  j>0 &&(i2<nx-1) &&(j<ny-1) ;
	const bool x_l_bound = (get_local_id(0)==0);
	const bool x_h_bound = ((get_local_id(0)+get_local_size(0))==(get_local_size(0)*2-1));
	const bool y_l_bound = (get_local_id(1)==0);
	const bool y_h_bound = (get_local_id(1)==(get_local_size(1)-1));

	//register for bottom and top planes
	//because of thread coarsening, we need to doulbe registers
	float bottom=0.0f,bottom2=0.0f,top=0.0f,top2=0.0f;
	float current=0.0f,current2=0.0f;

	//load data for bottom and current 
	if((i<nx) &&(j<ny))
	{
		bottom=A0[Index3D (nx, ny, i, j, 0)];
		current=A0[Index3D (nx, ny, i, j, 1)];
	}
	
	if((i2<nx) &&(j<ny))
	{
		bottom2=A0[Index3D (nx, ny, i2, j, 0)];
		current2=A0[Index3D (nx, ny, i2, j, 1)];
	}

	
	for(int k=1;k<nz-1;k++)
	{		
		//load required data on xy planes
		if((i<nx) &&(j<ny))
			top=A0[Index3D (nx, ny, i, j, k+1)];
		
		if(w_region)
		{		
			Anext[Index3D (nx, ny, i, j, k)] =
				(top + bottom + 
				A0[Index3D(nx,ny,i,j+1,k)] + 
				A0[Index3D(nx,ny,i,j-1,k)] +
				A0[Index3D (nx, ny, i+1, j, k )] +
				A0[Index3D (nx, ny, i-1, j, k )])*c1 -
				current*c0;
		}
	
		//load another block 
		if((i2<nx) &&(j<ny))
			top2=A0[Index3D (nx, ny, i2, j, k+1)];
		
		if(w_region2)
		{			      		
			Anext[Index3D (nx, ny, i2, j, k)] =
				(top2 + bottom2 + 
				A0[Index3D(nx,ny,i2,j+1,k)] + 
				A0[Index3D(nx,ny,i2,j-1,k)] + 
				A0[Index3D (nx, ny, i2+1, j, k )] +
				A0[Index3D (nx, ny, i2-1, j, k )])*c1 - 
				current2*c0;
		}

		//swap data

		bottom=current;
		current=top;
		bottom2=current2;
		current2=top2;
		
	}
}
