/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <parboil.h>

#include "args.h"
#include "model.h"

#include <visc.h>

extern unsigned int NUM_SETS;
extern unsigned int NUM_ELEMENTS;


#define WARP_SIZE 32
#define NUM_BANKS 16
#define LOG_NUM_BANKS 4

#define BLOCK_SIZE 256
#define NUM_WARPS (BLOCK_SIZE/WARP_SIZE)
#define HISTS_PER_WARP 16
#define NUM_HISTOGRAMS  (NUM_WARPS*HISTS_PER_WARP)
#define THREADS_PER_HIST (WARP_SIZE/HISTS_PER_WARP)

#define warp_hists(x,y) warp_hists[(x)*NUM_HISTOGRAMS+(y)]


typedef struct __attribute__((__packed__)) {
  hist_t* histograms;
  size_t bytes_histograms;
  float* all_x_data;
  size_t bytes_all_data;
  float* binb;
  size_t bytes_binb;
  int NUM_SETS;
  int NUM_ELEMENTS;
  long block;
  long grid;
}
RootIn;

void packData(  RootIn* args,
                hist_t* histograms,
                size_t bytes_histograms,
                float* all_x_data,
                size_t bytes_all_data,
                float* binb,
                size_t bytes_binb,
                int NUM_SETS,
                int NUM_ELEMENTS,
                long block,
                long grid) {
  args->histograms = histograms;
  args->bytes_histograms = bytes_histograms;
  args->all_x_data = all_x_data;
  args->bytes_all_data = bytes_all_data;
  args->binb = binb;
  args->bytes_binb = bytes_binb;
  args->NUM_SETS = NUM_SETS;
  args->NUM_ELEMENTS = NUM_ELEMENTS;
  args->block = block;
  args->grid = grid;
}


void Allocation(long block) {
  // Memory shared between threadblocks
  //void* data_s = __visc__malloc(sizeof(struct cartesian)*BLOCK_SIZE);
  void* warp_hists = __visc__malloc(sizeof(unsigned int)*NUM_BINS*NUM_HISTOGRAMS);

  //__visc__return(data_s, sizeof(struct cartesian)*BLOCK_SIZE,
                 //warp_hists, sizeof(unsigned int)*NUM_BINS*NUM_HISTOGRAMS);
  __visc__return(2, warp_hists, sizeof(unsigned int)*NUM_BINS*NUM_HISTOGRAMS);
}

void TPACFLeaf(hist_t* histograms, size_t bytes_histograms,
               // next two args are read-only arrays
               float* all_x_data, size_t bytes_all_data,
               float* binb, size_t bytes_binb,
               int NUM_SETS, int NUM_ELEMENTS,
               // shared memory args
               //struct cartesian* data_s, size_t bytes_data_s,
               unsigned int* warp_hists, size_t bytes_warp_hists) {

  __visc__hint(visc::DEVICE);
  __visc__attributes(2, all_x_data, binb, 1, histograms);

  void* thisNode = __visc__getNode();
  void* parentNode = __visc__getParentNode(thisNode);
  int lx = __visc__getNodeInstanceID_x(thisNode);
  int gx = __visc__getNodeInstanceID_x(parentNode);
  int dimx = __visc__getNumNodeInstances_x(thisNode);

  float* all_y_data = all_x_data + NUM_ELEMENTS*(NUM_SETS+1); 
  float* all_z_data = all_y_data + NUM_ELEMENTS*(NUM_SETS+1);

  unsigned int bx = gx;
  unsigned int tid = lx;
  bool do_self = (bx < (NUM_SETS + 1));

  float* data_x;
  float* data_y;
  float* data_z;
  float* random_x;
  float* random_y;
  float* random_z;

  for(unsigned int w = 0; w < NUM_BINS*NUM_HISTOGRAMS; w += BLOCK_SIZE )
    {
      if((w+tid) < (NUM_BINS*NUM_HISTOGRAMS))
	{
	  warp_hists((w+tid)/NUM_HISTOGRAMS, (w+tid)%NUM_HISTOGRAMS) = 0;
	}
    }

  // Get stuff into shared memory to kick off the loop.
  if( !do_self)
    {
      data_x = all_x_data;
      data_y = all_y_data;
      data_z = all_z_data;

      random_x = all_x_data + NUM_ELEMENTS * (bx - NUM_SETS);
      random_y = all_y_data + NUM_ELEMENTS * (bx - NUM_SETS);
      random_z = all_z_data + NUM_ELEMENTS * (bx - NUM_SETS);
    }
  else
    {
      random_x = all_x_data + NUM_ELEMENTS * (bx);
      random_y = all_y_data + NUM_ELEMENTS * (bx);
      random_z = all_z_data + NUM_ELEMENTS * (bx);
      
      data_x = random_x;
      data_y = random_y;
      data_z = random_z;
    }
 
  // Iterate over all random points
  for(unsigned int j = 0; j < NUM_ELEMENTS; j += BLOCK_SIZE)
    {
      // load current random point values
      float random_x_s;
      float random_y_s;
      float random_z_s;
	  
      if(tid + j < NUM_ELEMENTS)
        {
	  random_x_s = random_x[tid + j];
	  random_y_s = random_y[tid + j];
	  random_z_s = random_z[tid + j];
	}

      // Iterate over all data points
      // If do_self, then use a tighter bound on the number of data points.
      for(unsigned int k = 0;
	  k < NUM_ELEMENTS && (do_self ? k < j + BLOCK_SIZE : 1); k++)
	{
	  // do actual calculations on the values:
	  float distance = data_x[k] * random_x_s + 
	    data_y[k] * random_y_s + 
	    data_z[k] * random_z_s ;

	  unsigned int bin_index;

	  // run binary search to find bin_index
	  unsigned int min = 0;
	  unsigned int max = NUM_BINS;
	  {
	    unsigned int k2;
	      
	    while (max > min+1)
	      {
		k2 = (min + max) / 2;
		if (distance >= binb[k2]) 
		  max = k2;
		else 
		  min = k2;
	      }
	    bin_index = max - 1;
	  }

	  unsigned int warpnum = tid / (WARP_SIZE/HISTS_PER_WARP);
	  if((distance < binb[min]) && (distance >= binb[max]) && 
	     (!do_self || (tid + j > k)) && ((tid + j) < NUM_ELEMENTS))
	    {
	      __visc__atomic_add((int*)&(warp_hists(bin_index, warpnum)), 1);
	    }
	}
    }
    
  // coalesce the histograms in a block
  unsigned int warp_index = tid & ( (NUM_HISTOGRAMS>>1) - 1);
  unsigned int bin_index = tid / (NUM_HISTOGRAMS>>1);
  for(unsigned int offset = NUM_HISTOGRAMS >> 1; offset > 0; 
      offset >>= 1)
    {
      for(unsigned int bin_base = 0; bin_base < NUM_BINS; 
	  bin_base += BLOCK_SIZE/ (NUM_HISTOGRAMS>>1))
	{
	  __visc__barrier();
	  if(warp_index < offset && bin_base+bin_index < NUM_BINS )
	    {
	      unsigned long sum =
		warp_hists(bin_base + bin_index, warp_index) + 
		warp_hists(bin_base + bin_index, warp_index+offset);
	      warp_hists(bin_base + bin_index, warp_index) = sum;
	    }
	}
    }
    
  __visc__barrier();
    
  // Put the results back in the real histogram
  // warp_hists(x, 0) holds sum of all locations of bin x
  hist_t* hist_base = histograms + NUM_BINS * bx;
  if(tid < NUM_BINS)
    {
      hist_base[tid] = warp_hists(tid, 0);
    }
}

void BlockingTPACF(hist_t* histograms, size_t bytes_histograms,
                   float* all_x_data, size_t bytes_all_data,
                   // next arg is read-only constant
                   float* binb, size_t bytes_binb,
                   int NUM_SETS, int NUM_ELEMENTS,
                   long block,
                   // shared memory args
                   struct cartesian* data_s, size_t bytes_data_s,
                   unsigned int* warp_hists, size_t bytes_warp_hists) {

  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, all_x_data, binb, 1, histograms);

  void* AllocationNode = __visc__createNodeND(0, Allocation);
  void* TPACFLeafNode = __visc__createNodeND(1, TPACFLeaf, block);

  // Bind Inputs
  __visc__bindIn(AllocationNode, 8, 0, 0); // Bind block
  __visc__bindIn(TPACFLeafNode, 0, 0, 0); // Bind histograms
  __visc__bindIn(TPACFLeafNode, 1, 1, 0); // Bind bytes_histograms
  __visc__bindIn(TPACFLeafNode, 2, 2, 0); // Bind all_x_data
  __visc__bindIn(TPACFLeafNode, 3, 3, 0); // Bind bytes_all_data
  __visc__bindIn(TPACFLeafNode, 4, 4, 0); // Bind binb
  __visc__bindIn(TPACFLeafNode, 5, 5, 0); // Bind bytes_binb
  __visc__bindIn(TPACFLeafNode, 6, 6, 0); // Bind NUM_SETS
  __visc__bindIn(TPACFLeafNode, 7, 7, 0); // Bind NUM_ELEMENTS

  // Create Edges
  __visc__edge(AllocationNode, TPACFLeafNode, 1, 0, 8, 0); // Edge warp_hists
  __visc__edge(AllocationNode, TPACFLeafNode, 1, 1, 9, 0); // Edge bytes_warp_hists

}

void TPACFRoot(hist_t* histograms, size_t bytes_histograms,
               float* all_x_data, size_t bytes_all_data,
               // next arg is read-only constant
               float* binb, size_t bytes_binb,
               int NUM_SETS, int NUM_ELEMENTS,
               long block,
               long grid) {

  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, all_x_data, binb, 1, histograms);

  void* BlockingTPACFNode = __visc__createNodeND(1, BlockingTPACF, grid);

  // Bind Inputs
  __visc__bindIn(BlockingTPACFNode, 0, 0, 0); // Bind histograms
  __visc__bindIn(BlockingTPACFNode, 1, 1, 0); // Bind bytes_histograms
  __visc__bindIn(BlockingTPACFNode, 2, 2, 0); // Bind all_x_data
  __visc__bindIn(BlockingTPACFNode, 3, 3, 0); // Bind bytes_all_data
  __visc__bindIn(BlockingTPACFNode, 4, 4, 0); // Bind binb
  __visc__bindIn(BlockingTPACFNode, 5, 5, 0); // Bind bytes_binb
  __visc__bindIn(BlockingTPACFNode, 6, 6, 0); // Bind NUM_SETS
  __visc__bindIn(BlockingTPACFNode, 7, 7, 0); // Bind NUM_ELEMENTS
  __visc__bindIn(BlockingTPACFNode, 8, 8, 0); // Bind block

}

// **===-----------------------------------------------------------===**

int
main( int argc, char** argv)
{
  struct pb_TimerSet timers;
  struct pb_Parameters* params;

  params = pb_ReadParameters( &argc, argv );

  options args;
  parse_args(argc, argv, &args);


  NUM_ELEMENTS = args.npoints;
  NUM_SETS = args.random_count;
  int num_elements = NUM_ELEMENTS;

  printf("Min distance: %f arcmin\n", min_arcmin);
  printf("Max distance: %f arcmin\n", max_arcmin);
  printf("Bins per dec: %i\n", bins_per_dec);
  printf("Total bins  : %i\n", NUM_BINS);

  //read in files
  unsigned mem_size = (1+NUM_SETS)*num_elements*sizeof(struct cartesian);
  unsigned f_mem_size = (1+NUM_SETS)*num_elements*sizeof(float);

  // container for all the points read from files
  struct cartesian *h_all_data;
  h_all_data = (struct cartesian*) malloc(mem_size);
  // Until I can get libs fixed

  // iterator for data files
  struct cartesian *working = h_all_data;


  // go through and read all data and random points into h_all_data
  //pb_SwitchToTimer( &timers, pb_TimerID_IO );
  readdatafile(params->inpFiles[0], working, num_elements);

  //pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );

  working += num_elements;
  for(int i = 0; i < (NUM_SETS); i++)
  {
    //pb_SwitchToTimer( &timers, pb_TimerID_IO );
    readdatafile(params->inpFiles[i+1], working, num_elements);
    //pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );

    working += num_elements;
  }

  pb_InitializeTimerSet( &timers );
  __visc__init();

  pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
  // split into x, y, and z arrays
  // AOS to SOA transformation
  size_t bytes_h_x_data = 3*f_mem_size;
  float * h_x_data = (float*) malloc (bytes_h_x_data);
  llvm_visc_track_mem(h_x_data, bytes_h_x_data);

  float * h_y_data = h_x_data + NUM_ELEMENTS*(NUM_SETS+1);
  float * h_z_data = h_y_data + NUM_ELEMENTS*(NUM_SETS+1);
  for(int i = 0; i < (NUM_SETS+1); ++i)
  {
    for(int j = 0; j < NUM_ELEMENTS; ++j)
    {
      h_x_data[i*NUM_ELEMENTS+j] = h_all_data[i*NUM_ELEMENTS+j].x;
      h_y_data[i*NUM_ELEMENTS+j] = h_all_data[i*NUM_ELEMENTS+j].y;
      h_z_data[i*NUM_ELEMENTS+j] = h_all_data[i*NUM_ELEMENTS+j].z;
    }
  }

  // from on use x, y, and z arrays, free h_all_data
  free(h_all_data);

  // allocate system memory for final histograms
  size_t bytes_hists = NUM_BINS*(NUM_SETS*2+1)*sizeof(hist_t);
  hist_t *hists = (hist_t *) malloc(bytes_hists);
  llvm_visc_track_mem(hists, bytes_hists);

  // Initialize the boundary constants for bin search
  size_t bytes_binb = (NUM_BINS+1)*sizeof(float);
  float *binb = (float*)malloc(bytes_binb);
  llvm_visc_track_mem(binb, bytes_binb);

  for (int k = 0; k < NUM_BINS+1; k++)
  {
    binb[k] = cos(pow(10.0, (log10(min_arcmin) + k*1.0/bins_per_dec))
                  / 60.0*D2R);
  }

  // **===------------------ Kick off TPACF on OpenCL------------------===**

  long block = BLOCK_SIZE;
  long grid = (NUM_SETS*2 + 1);

  RootIn* graph_args = (RootIn*) malloc (sizeof(RootIn));
  packData(graph_args,
           hists,
           bytes_hists,
           h_x_data,
           bytes_h_x_data,
           binb,
           bytes_binb,
           NUM_SETS,
           NUM_ELEMENTS,
           block,
           grid);
  pb_SwitchToTimer( &timers, visc_TimerID_COMPUTATION );
  
  void* TPACF_DFG = __visc__launch(0, TPACFRoot, (void*)graph_args);
  __visc__wait(TPACF_DFG);

  pb_SwitchToTimer( &timers, pb_TimerID_COPY );

  pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
  // **===-------------------------------------------------------------===**

  llvm_visc_request_mem(hists, bytes_hists);
  // references into output histograms
  hist_t *dd_hist = hists;
  hist_t *rr_hist = dd_hist + NUM_BINS;
  hist_t *dr_hist = rr_hist + NUM_BINS*NUM_SETS;

  // add up values within dr and rr
  int rr[NUM_BINS];
  for(int i=0; i<NUM_BINS; i++)
  {
    rr[i] = 0;
  }
  for(int i=0; i<NUM_SETS; i++)
  {
    for(int j=0; j<NUM_BINS; j++)
    {
      rr[j] += rr_hist[i*NUM_BINS + j];
    }
  }
  int dr[NUM_BINS];
  for(int i=0; i<NUM_BINS; i++)
  {
    dr[i] = 0;
  }
  for(int i=0; i<NUM_SETS; i++)
  {
    for(int j=0; j<NUM_BINS; j++)
    {
      dr[j] += dr_hist[i*NUM_BINS + j];
    }
  }

  pb_SwitchToTimer(&timers, pb_TimerID_NONE);
  pb_PrintTimerSet(&timers);
  __visc__cleanup();

  FILE *outfile;
  if ((outfile = fopen(params->outFile, "w")) == NULL)
  {
    fprintf(stderr, "Unable to open output file %s for writing, "
            "assuming stdout\n", params->outFile);
    outfile = stdout;
  }

  //pb_SwitchToTimer( &timers, pb_TimerID_IO );
  // print out final histograms + omega (while calculating omega)
  for(int i=0; i<NUM_BINS; i++)
  {
    fprintf(outfile, "%d\n%d\n%d\n", dd_hist[i], dr[i], rr[i]);
  }

  //pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
  if(outfile != stdout)
    fclose(outfile);

  // cleanup memory
  free(hists);
  free(h_x_data);

}

